/*
Copyright (C) 2012-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// buffers.c - data buffers


#include "core.h"
#include "corex.h" // We are not part of core
#include "buffers.h" // Courtesy include

/*
typedef struct sizebuf_s
{
	cbool	allowoverflow;	// if false, do a System_Error
	cbool	overflowed;		// set to true if the buffer size failed
	byte	*data;
	int		maxsize;
	int		cursize;
} sizebuf_t;
*/


///////////////////////////////////////////////////////////////////////////////
//  Buffers: cmd.c, network and server use these functions to fill buffers
///////////////////////////////////////////////////////////////////////////////

void Buffer_Alloc (sizebuf_t *buf, int len)
{
	buf->data = (byte *) core_calloc (len, 1);
	buf->maxsize = len;
	buf->cursize = 0;
}


void Buffer_Clear (sizebuf_t *buf)
{
	buf->cursize = 0;
}


void Buffer_Free (sizebuf_t *buf)
{
	core_free (buf->data);
}


void *Buffer_GetSpace (sizebuf_t *buf, int len)
{
	void *data;

	if (buf->cursize + len > buf->maxsize)
	{
		if (!buf->allowoverflow)
			Core_Error ("SZ_GetSpace: overflow without allowoverflow set");

		if (len > buf->maxsize)
			Core_Error ("SZ_GetSpace: %i is > full buffer size", len);

		buf->overflowed = true;
		Core_Printf ("SZ_GetSpace: overflow\n");
		Buffer_Clear (buf);
	}

	data = buf->data + buf->cursize;
	buf->cursize += len;

	return data;
}

void Buffer_Write (sizebuf_t *buf, const void *data, int len)
{
	memcpy (Buffer_GetSpace(buf, len), data, len);
}

void Buffer_Print (sizebuf_t *buf, const char *s)
{
	cbool trailing_null = buf->cursize ? !buf->data[buf->cursize - 1] : false;
	int len = strlen (s) + 1;

	switch (trailing_null) 
	{
	case 0:
		// write over trailing 0, get one less byte
		memcpy ((byte *)Buffer_GetSpace(buf, len - 1) - 1, s, len);
		break;
	
	default:
		// no trailing 0
		memcpy ((byte *)Buffer_GetSpace(buf, len), s, len);
		break;

	}
}


cbool Buffer_GetLine (sizebuf_t *buf, char *line, int len)
{
	if (!buf->cursize)
		return false;
	else
	{
	// find a \n or ; line break
		int i;
		int		quotes;		
		char	*text = (char *)buf->data;

		for (i = 0, quotes = 0 ; i < buf->cursize ; i++)
		{
			if (text[i] == '"')
				quotes++;
			
			if ( !(quotes & 1) &&  text[i] == ';')
				break;	// don't break if inside a quoted string

			if (text[i] == '\n')
				break;
		}

		// Found a line
		strlcpy (line, text, c_min(len, i + 1 /* +1 = space for NULL */) ); // Baker: copy the lesser of len or i
		line[i] = 0;

	// delete the text from the command buffer and move remaining commands down
	// this is necessary because commands (exec, alias) can insert data at the
	// beginning of the text buffer

		if (i == buf->cursize)
			buf->cursize = 0; // End of buffer, reset 
		else
		{
			i++;
			buf->cursize -= i;
			memmove (text, text + i, buf->cursize); // Move it up
		}

		return true;
	}
	
}

