/*
Copyright (C) 2012-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// buffers.h - data buffers

#ifndef __BUFFERS_H__
#define __BUFFERS_H__

///////////////////////////////////////////////////////////////////////////////
//  Data buffers ... to fill and be emptied in command and communications
///////////////////////////////////////////////////////////////////////////////

typedef struct sizebuf_s
{
	cbool	allowoverflow;	// if false, do a Core_Error
	cbool	overflowed;		// set to true if the buffer size failed
	byte	*data;
	int		maxsize;
	int		cursize;
} sizebuf_t;

void Buffer_Alloc (sizebuf_t *buf, int len); // dynamic allocation based on size
void Buffer_Clear (sizebuf_t *buf); // set size to 0
void Buffer_Free (sizebuf_t *buf); // de-allocate

void *Buffer_GetSpace (sizebuf_t *buf, int len);
void Buffer_Write (sizebuf_t *buf, const void *data, int len); // Copies data into buffer
void Buffer_Print (sizebuf_t *buf, const char *data);	// strcats into the buffer
cbool Buffer_GetLine (sizebuf_t *buf, char *line, int len); // Obtains a line, stops at semi-colons or newlines, returns false if empty buffer


#endif // __BUFFERS_H__



