/*
Copyright (C) 2012-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
// console.c -- memory chain for temp allocs
#if 0
#include "core.h"
#include "memchain.h"

///////////////////////////////////////////////////////////////////////////////
//  LOCAL SETUP
///////////////////////////////////////////////////////////////////////////////

#define MLOCAL(_m) struct mlocal *_m = (struct mlocal *) me->_local;

// Begin ...

static const char * const _tag = "console";  	// TAG
#define mobj_t console_t						// object type
#define mlocal _console_local					// private data struct

typedef struct
{
	char		*text;					// Pointer within buffer to current line
	
	int			column;
	int			selection_length;		// If non-zero, there is a selection
	
	int			completion_type;		// User defined
	int			completion_start;		// Start of current word
	const char	*completion;			// Current auto-complete
} entry_box_t;

// History?
#define MAX_HISTORY_64

struct mlocal							// private data
{
	// Make this
	memchain_t *memchain;
	char		history[MAX_HISTORY_64][MAX_CMD_256];

	char		*buffer;				// Text allocation
	int			bufsize;

	int			rows;
	int			cols;

	int			row;
	entry_box_t entrybox;

// Display
	int			backscroll;				// Are we viewing current line
} console_t;


///////////////////////////////////////////////////////////////////////////////
//  INTERNAL FUNCTIONS
///////////////////////////////////////////////////////////////////////////////

static void NewLine (mobj_t *me)
{
	MLOCAL(m)

	memset (m->entrybox, 0, sizeof(m->entrybox));
	m->entrybox.text = where?
	m->backscroll = 0;
}

static void Clear (mobj_t *me)
{
	MLOCAL(m)

	// Erase buffer
	memset (m->buffer, 0, bufsize);
	
	// What else?
	m->backscroll = 0;
}


/*
================
Con_ToggleConsole_f
================
*/
extern int history_line; //johnfitz

// Exiting the console ... set things.
void Con_Exit (void)
{
	History_Save (); // Baker: Keep history up-to-date.
	key_lines[edit_line][1] = 0;	// clear any typing
	key_linepos = 1;
	console1.backscroll = 0; //johnfitz -- toggleconsole should return you to the bottom of the scrollback
	history_line = edit_line; //johnfitz -- it should also return you to the bottom of the command history
}


/*
================
Con_Clear_f
================
*/
void Con_Clear_f (void)
{
	if (console1.text)
		memset (console1.text, ' ', console1.buffer_size); //johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	console1.backscroll = 0; //johnfitz -- if console is empty, being scrolled up is confusing
}

/*
================
Con_Dump_f -- johnfitz -- adapted from quake2 source
================
*/
void Con_Dump_f (void)
{
	int		l, x;
	const char	*line;
	FILE	*f;
	char	buffer[1024];
	char	name[MAX_OSPATH];

#if 1
	//johnfitz -- there is a security risk in writing files with an arbitrary filename. so,
	//until stuffcmd is crippled to alleviate this risk, just force the default filename.
	FS_FullPath_From_QPath (name, "condump.txt");
#else
	if (line->count > 2)
	{
		Con_Printf ("usage: condump <filename>\n");
		return;
	}

	if (line->count > 1)
	{
		if (strstr(Cmd_Argv(1), ".."))
		{
			Con_Printf ("Relative pathnames are not allowed.\n");
			return;
		}
		FS_FullPath_From_QPath (name, Cmd_Argv(1));
		File_URL_Edit_Force_Extension (name, ".txt", sizeof(name));
	}
	else
		FS_FullPath_From_QPath (name, "condump.txt");
#endif

	f = FS_fopen_write_create_path (name, "w");
	if (!f)
	{
		Con_Printf ("ERROR: couldn't open file %s.\n", name);
		return;
	}

	// skip initial empty lines
	for (l = console1.cursor_row - console1.buffer_rows + 1 ; l <= console1.cursor_row ; l++)
	{
		line = console1.text + (l%console1.buffer_rows)*console1.buffer_columns;
		for (x=0 ; x<console1.buffer_columns ; x++)
			if (line[x] != ' ')
				break;
		if (x != console1.buffer_columns)
			break;
	}

	// write the remaining lines
	buffer[console1.buffer_columns] = 0;
	for ( ; l <= console1.cursor_row ; l++)
	{
		line = console1.text + (l%console1.buffer_rows)*console1.buffer_columns;
		strncpy (buffer, line, console1.buffer_columns);
		for (x = console1.buffer_columns - 1 ; x >= 0 ; x--)
		{
			if (buffer[x] == ' ')
				buffer[x] = 0;
			else
				break;
		}
		for (x=0; buffer[x]; x++)
			buffer[x] &= 0x7f;

		fprintf (f, "%s\n", buffer);
	}

	FS_fclose (f);
	Recent_File_Set_QPath (name);
	Con_Printf ("Dumped console text to %s.\n", name);
}

/*
================
Con_Copy_f -- Baker -- adapted from Con_Dump
================
*/
void Con_Copy_f (lparse_t *line)
{
	char	outstring[CONSOLE_TEXTSIZE]="";
	int		l, x;
//	char	*line;
	char	buffer[1024];

	if (line->count > 1) // More than just the command
	{
#ifdef SUPPORTS_WINDOWS_MENUS // Baker change
	// To be able to put this info on clipboard in menu
		if (strcasecmp (Cmd_Argv(1), "vp80") == 0)
		{
			System_CopyToClipboard("http://www.optimasc.com/products/vp8vfw/index.html");
			Con_Printf ("Copied Google VP80/WebM Codec URL to clipboard.\n");
			return;
		}
#endif // Baker change +

		if (strcasecmp (Cmd_Argv(1), "ents") == 0)
		{
			extern char *entity_string;
			if (!sv.active || !entity_string)
			{
				Con_Printf ("copy ents: Not running a server or no entities");
				return;
			}

			Clipboard_Set_Text (entity_string);

			Con_Printf ("Entities copied to the clipboard (%i bytes)\n", (int)strlen(entity_string));
			return;
		}

		if (strcasecmp (Cmd_Argv(1), "screen") == 0)
		{
			SCR_ScreenShot_Clipboard_f ();
			return;
		}

		// Invalid args
		Con_Printf ("Usage: %s [ents | screen]: copies console to clipboard\n\"copy ents\" copies map entities to clipboard.\n\"copy screen\" copies screen to clipboard.", Cmd_Argv (0));
		return;
	}

	// skip initial empty lines
	for (l = console1.cursor_row - console1.buffer_rows + 1 ; l <= console1.cursor_row ; l++)
	{
		char	*line  = console1.text + (l%console1.buffer_rows)*console1.buffer_columns;
		for (x = 0 ; x < console1.buffer_columns ; x++)
			if (line[x] != ' ')
				break;
		if (x != console1.buffer_columns)
			break;
	}

	// write the remaining lines
	buffer[console1.buffer_columns] = 0;
	for ( ; l <= console1.cursor_row ; l++)
	{
		char *line = console1.text + (l%console1.buffer_rows)*console1.buffer_columns;
		strncpy (buffer, line, console1.buffer_columns);
		for (x = console1.buffer_columns - 1 ; x >= 0 ; x--)
		{
			if (buffer[x] == ' ')
				buffer[x] = 0;
			else
				break;
		}
		for (x=0; buffer[x]; x++)
			buffer[x] &= 0x7f;

		c_strlcat (outstring, va("%s\r\n", buffer));

	}

	Clipboard_Set_Text (outstring);
	Con_Printf ("Copied console to clipboard\n");
}




/*
================
Con_CheckResize

If the line width has changed, reformat the buffer.
================
*/
void Con_CheckResize (void)
{
	int		i, j, width, oldwidth, oldtotallines, numlines, numchars;
	char	*tbuf; //johnfitz -- tbuf no longer a static array
	int mark; //johnfitz

	width = (vid.conwidth >> 3) - 2; //johnfitz -- use vid.conwidth instead of vid.width

	if (width == console1.buffer_columns)
		return;

	oldwidth = console1.buffer_columns;
	console1.buffer_columns = width;
	oldtotallines = console1.buffer_rows;
	console1.buffer_rows = console1.buffer_size / console1.buffer_columns; //johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	numlines = oldtotallines;

	if (console1.buffer_rows < numlines)
		numlines = console1.buffer_rows;

	numchars = oldwidth;

	if (console1.buffer_columns < numchars)
		numchars = console1.buffer_columns;

	mark = Hunk_LowMark (); //johnfitz
	tbuf = (char *) Hunk_Alloc (console1.buffer_size); //johnfitz

	memcpy (tbuf, console1.text, console1.buffer_size);//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	memset (console1.text, ' ', console1.buffer_size);//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE

	for (i=0 ; i<numlines ; i++)
	{
		for (j=0 ; j<numchars ; j++)
		{
			console1.text[(console1.buffer_rows - 1 - i) * console1.buffer_columns + j] =
					tbuf[((console1.cursor_row - i + oldtotallines) % oldtotallines) * oldwidth + j];
		}
	}

	Hunk_FreeToLowMark (mark); //johnfitz

	Con_ClearNotify ();

	console1.backscroll = 0;
	console1.cursor_row = console1.buffer_rows - 1;
}


/*
================
Con_Init
================
*/
void Con_Init (void)
{
	int i;
	char	temp[SYSTEM_STRING_SIZE];
	const char	*t2 = "qconsole.log";

#pragma message ("Baker: OH fucking groovy.  No -condebug for dedicated since doesn't reach here.")
	con_debuglog = COM_CheckParm("-condebug");

	if (con_debuglog)
	{
		FS_FullPath_From_QPath (temp, t2);
		File_Delete (temp);
	}

	if (isDedicated)
		return;

	//johnfitz -- user settable console buffer size
	i = COM_CheckParm("-consize");
	if (i && i < com_argc-1)
		console1.buffer_size = c_max(CONSOLE_MINSIZE, atoi(com_argv[i+1]) * 1024);
	else
		console1.buffer_size = CONSOLE_TEXTSIZE;
	//johnfitz

	console1.text = (char *) Hunk_AllocName (console1.buffer_size, "context");//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	memset (console1.text, ' ', console1.buffer_size);//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	console1.buffer_columns = -1;

	//johnfitz -- no need to run Con_CheckResize here
	console1.buffer_columns = 78;
	console1.buffer_rows = console1.buffer_size / console1.buffer_columns;//johnfitz -- console1.buffer_size replaces CON_TEXTSIZE
	console1.backscroll = 0;
	console1.cursor_row = console1.buffer_rows - 1;
	console1.user_pct = 0.50; // Default console is 50%
	//johnfitz

	Con_SafePrintf ("Console initialized.\n");


	Cmd_AddCommands (Con_Init);

	console1.initialized = true;
}


/*
===============
Con_Linefeed
===============
*/
static void Con_Linefeed (void)
{
	//johnfitz -- improved scrolling
	if (console1.backscroll)
		console1.backscroll++;
	if (console1.backscroll > console1.buffer_rows - (clheight >> 3) - 1)
		console1.backscroll = console1.buffer_rows - (clheight >> 3) - 1;
	//johnfitz

	console1.cursor_column = 0;
	console1.cursor_row++;
	memset (&console1.text[(console1.cursor_row%console1.buffer_rows)*console1.buffer_columns], ' ', console1.buffer_columns);
}

/*
================
Con_Print

Handles cursor positioning, line wrapping, etc
All console printing must go through this in order to be logged to disk
If no console is visible, the notify window will pop up.
================
*/
static void Con_Print (const char *txt)
{
	int		y;
	int		c, l;
	static int	cr;
	int		mask;


	//console1.backscroll = 0; //johnfitz -- better console scrolling

	if (txt[0] == 1)
	{
		mask = 128;		// go to colored text
		S_LocalSound ("misc/talk.wav"); // play talk wav
		txt++;
	}
	else if (txt[0] == 2)
	{
		mask = 128;		// go to colored text
		txt++;
	}
	else
	{
		mask = 0;
	}

	while ( (c = *txt) )
	{
	// count word length
		for (l = 0 ; l < console1.buffer_columns ; l++)
			if ( txt[l] <= ' ')
				break;

	// word wrap
		if (l != console1.buffer_columns && (console1.cursor_column + l > console1.buffer_columns) )
			console1.cursor_column = 0;

		txt++;

		if (cr)
		{
			console1.cursor_row--;
			cr = false;
		}

		if (!console1.cursor_column)
		{
			Con_Linefeed ();
		// mark time for transparent overlay
			if (console1.cursor_row >= 0)
				console1.con_times[console1.cursor_row % (int)CONSOLE_NOTIFY_TIMES] = realtime;
		}

		switch (c)
		{
		case '\n':
			console1.cursor_column = 0;
			break;

		case '\r':
			console1.cursor_column = 0;
			cr = 1;
			break;

		default:	// display character and advance
			y = console1.cursor_row % console1.buffer_rows;
			console1.text[y*console1.buffer_columns+console1.cursor_column] = c | mask;
			console1.cursor_column++;
			if (console1.cursor_column >= console1.buffer_columns)
				console1.cursor_column = 0;
			break;
		}
	}
}



/*
================
Con_DrawInput -- johnfitz -- modified to allow insert editing

The input line scrolls horizontally if typing goes beyond the right edge
================
*/
extern qpic_t *pic_ovr, *pic_ins; //johnfitz -- new cursor handling

void Con_DrawInput (void)
{
	int		i, len, ofs = 0;
	char	c[CONSOLE_MAX_CMDLINE], *text;

	if (key_dest != key_console && !console1.forcedup)
		return;		// don't draw anything

	c_strlcpy (c, key_lines[edit_line]);
	text = c;

// pad with one space so we can draw one past the string (in case the cursor is there)
	len = strlen(key_lines[edit_line]);
	text[len] = ' ';
	text[len+1] = 0;

// prestep if horizontally scrolling
	if (key_linepos >= console1.buffer_columns)
	{
		ofs = 1 + key_linepos - console1.buffer_columns;
		text += 1 + key_linepos - console1.buffer_columns;
	}

// draw input string
#pragma message ("Baker: I don't like this")
	for (i = 0; i <= (int)(strlen(text) - 1) && i < console1.buffer_columns; i++) //only write enough letters to go from *text to cursor
	{
		const char *xmatch_start = (match_start - key_lines[edit_line]) + c;
		const char *xmatch_end = xmatch_start + strlen(key_tabpartial) - 1;
//		char *last_char_before_cursor = key_lines[edit_line] + key_linepos - 1;
		if (key_tabpartial[0] && xmatch_start <= &text[i] &&  &text[i] <= xmatch_end)
		{
			Draw_Character ((i+1)<<3, vid.conheight - 16, text[i] | 128); // Bronze tab completion in progress
		}
		else Draw_Character ((i+1)<<3, vid.conheight - 16, text[i]);

	}

// johnfitz -- new cursor handling
	if (!((int)((realtime-key_blinktime) * CONSOLE_CURSOR_SPEED) & 1))
#ifdef GLQUAKE // Baker: Eliminate in future
		Draw_Pic ((key_linepos + 1 - ofs)<<3, vid.conheight - 16, key_insert ? pic_ins : pic_ovr);
#else
		if (console1.visible_lines > 16)
			Draw_Fill ((key_linepos+1 - ofs) << 3, vid.conheight - 16, key_insert ? 2 : 8, 8, 254, 1);
#endif // GLQUAKE vs. WINQUAKE
}

/*
================
Con_DrawConsole -- johnfitz -- heavy revision

Draws the console with the solid background
The typing input line at the bottom should only be drawn if typing is allowed
================
*/
void Con_DrawConsole (float pct, cbool drawinput)
{
	int				i, j, x, y, rows;
	int				sb;

	const char	*text;
	char			ver[32];

	if (pct <= 0)
		return;

// draw the background
	Draw_SetCanvas (CANVAS_CONSOLE);
	Draw_ConsoleBackground ();


	// Baker: Don't draw the console text if menu is being drawn.
	if (key_dest == key_menu && m_state)
		return;

// draw the text
	rows = (console1.visible_lines_conscale + 7)/8;
	y = vid.conheight - rows * 8;
	rows -= 2; //for input and version lines
	sb = (console1.backscroll) ? 2 : 0;

	for (i = console1.cursor_row - rows + 1 ; i <= console1.cursor_row - sb ; i++, y+=8)
	{
		j = i - console1.backscroll;
		if (j<0)
			j = 0;
		text = console1.text + (j % console1.buffer_rows) * console1.buffer_columns;

		for (x=0 ; x < console1.buffer_columns ; x++)
			Draw_Character ( (x+1)<<3, y, text[x]);
	}

// draw scrollback arrows
	if (console1.backscroll)
	{
		y+=8; // blank line
		for (x=0 ; x < console1.buffer_columns ; x+=4)
			Draw_Character ((x+1)<<3, y, '^');
		y+=8;
	}

// draw the input prompt, user text, and cursor
	if (drawinput)
		Con_DrawInput ();

//draw version number in bottom right
	y+=8;
	c_snprintf2 (ver, "%s %1.2f"/*" beta2"*/, ENGINE_NAME, (float)ENGINE_VERSION);
	for (x=0; x < (int) strlen(ver); x++)
		Draw_Character ((console1.buffer_columns-strlen(ver)+x+2)<<3, y, ver[x] /*+ 128*/);
}


#endif