/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2009-2014 Baker and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// image.h -- image manipulation and loading

#ifndef __IMAGE_H__
#define __IMAGE_H__

#define CORE_LOCAL
#include "core.h"
#include "environment.h"
#include <stdio.h> // fopen, etc.
#include <stdlib.h> // malloc, etc.
#include <string.h> // mem
#include "lodepng.h"


///////////////////////////////////////////////////////////////////////////////
//  IMAGE: OPERATIONS
///////////////////////////////////////////////////////////////////////////////

void Image_Flip_Buffer (byte *buffer, int columns, int rows, int bytes_per_pixel);

///////////////////////////////////////////////////////////////////////////////
//  IMAGE: MISCELLANEOUS
///////////////////////////////////////////////////////////////////////////////

int Image_Power_Of_Two_Size (int myInteger);


///////////////////////////////////////////////////////////////////////////////
//  IMAGE: GAMMA
///////////////////////////////////////////////////////////////////////////////

void Image_Build_Gamma_Table (float g, float c, byte gammatable_256[]);
void Image_Apply_Gamma_To_RGBA_Pixels (byte *data, int width, int height, byte gammatable_256[]);

///////////////////////////////////////////////////////////////////////////////
//  PALETTE REPLACEMENT
///////////////////////////////////////////////////////////////////////////////

// 256 color functions
void Image_Quick_Palette_256_Free (void);
void Image_Quick_Palette_256_Alloc (int black_index);
byte Palette_Best_Color_Index_For_Pixel (unsigned mypixel, byte palette768[], int numpalcolors);
void Image_Convert_Palette_To_Palette (byte *pixels_indexed, int len, byte old_palette_768[], byte new_palette_768[] );
byte *Image_Buffer_RGBA_To_Palette_Alpha_Threshold_Alloc (byte *pixels_rgba, int width, int height, byte palette_768[], int alpha_threshold);

// RGB functions
extern unsigned int *fastpal_rgb4;
void Image_Quick_Palette_RGBA_Free (void);
void Image_Quick_Palette_RGBA_Alloc (void);

///////////////////////////////////////////////////////////////////////////////
//  IMAGE: PNG, PCX (READ ONLY), TGA -- RGBA input and output
///////////////////////////////////////////////////////////////////////////////

byte *Image_Load_TGA_FileHandle (FILE *fin, int *width, int *height, malloc_fn_t Malloc_Fn, const char *description);
byte *Image_Load_PCX_FileHandle (FILE *fin, int *width, int *height, size_t filelen, malloc_fn_t Malloc_Fn, const char *description);

cbool Image_Save_TGA (const char *path_to_file, byte *data, int width, int height, int bits_per_pixel, cbool upsidedown);
cbool Image_Save_PNG (const char *path_to_file, byte *data, int width, int height);




#endif	// ! __IMAGE_H__

