/*
Copyright (C) 2014-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// net.h -- De-platformize network code


#ifndef __NET_H__
#define __NET_H__


#include "environment.h"

#ifdef PLATFORM_WINDOWS

	#include <winsock2.h> // Must be first
	#include <ws2tcpip.h>
	
	#include "core_windows.h"

	#ifdef __WINDOWS_VC6_BANDAGES__
		#ifndef IPV6_V6ONLY 
			#define IPV6_V6ONLY 27 
		#endif 

	#if 0
		__inline  void IN6_SET_ADDR_UNSPECIFIED(PIN6_ADDR a)
		{
			memset(a->u.Byte, 0, sizeof(IN6_ADDR));
		}

		__inline void IN6ADDR_SETANY(PSOCKADDR_IN6 a)
		{
			a->sin6_family = 23;
			a->sin6_port = 0;
			a->sin6_flowinfo = 0;
			IN6_SET_ADDR_UNSPECIFIED(&a->sin6_addr);
			a->sin6_scope_id = 0;
		}
	#endif	

		__inline void IN4ADDR_SETANY(PSOCKADDR_IN a)
		{
			a->sin_family = AF_INET;
			a->sin_port = 0;
			a->sin_addr.S_un.S_addr = (ULONG)0x00000000;
			memset(a->sin_zero, 0, sizeof(a->sin_zero));
		}
		
		__inline void INETADDR_SETANY(PSOCKADDR a)
		{
			if (a->sa_family == AF_INET6) {
				IN6ADDR_SETANY((PSOCKADDR_IN6)a);
			} else {
				((void) 0);
				IN4ADDR_SETANY((PSOCKADDR_IN)a);
			}
		}
		
		const char *inet_ntop(int af, const void *src, char *dst, socklen_t size);
	#endif // __WINDOWS_VC6_BANDAGES__



	#define sys_errno			WSAGetLastError()
	#define sys_closesocket		closesocket // Windows cannot use "close" to close sockets
	#define sys_sock_t			SOCKET
	#define sys_ioctlsocket		ioctlsocket	

	/* there is no in_addr_t on windows: define it as
	   the type of the S_addr of in_addr structure */
	//typedef u_long	in_addr_t;	/* uint32_t */
	typedef unsigned int in_addr_t;	/* uint32_t */

#else // ! PLATFORM_WINDOWS ...

	#define sys_errno			errno
	#define sys_closesocket		close	// BSD sockets use "close" to close sockets
	#define sys_sock_t			int
	#define sys_ioctlsocket		ioctl

#endif // PLATFORM_WINDOWS



void Net_Startup (void);
void Net_Cleanup (void);
const char *Net_StrError (int err); // Because I'm not seeing strerror translate error codes for WSA
cbool Net_Error (const char *fmt, ...);

// Buffer must be INET_ADDRSTRLEN length
void sockaddr_storage_printf (struct sockaddr_storage *sas);
void sockaddr_storage_get (struct sockaddr_storage *sas, char *addr, int *port);
void sockaddr_storage_set (struct sockaddr_storage *sas, const char *addr, int *port);
void sockaddr_storage_anyset (struct sockaddr_storage *sas, int *port);

int getaddrfamily(const char *addr);
const char *getaddrinfo_str_alloc (const char *ip_address, int port);

int inet_pton(int af, const char *src, void *dst);


#endif // ! __NET_H__

