/*
Copyright (C) 2014-2014 Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// netcon.h -- network connection object



#include "core.h"
#include "corex.h"




//#include <winsock2.h>


#pragma comment (lib, "wsock32.lib")
#pragma comment(lib,"ws2_32.lib")


#include "net.h"
#include "netcon.h" // courtesy include

///////////////////////////////////////////////////////////////////////////////
//  LOCAL SETUP
///////////////////////////////////////////////////////////////////////////////

#define MLOCAL(_m) struct mlocal *_m = (struct mlocal *) me->_local;

// Begin ...

static const char * const _tag = "netserv";  	// TAG
#define mobj_t netserv_t						// object type
#define mlocal _netserv_local					// private data struct

/*
typedef struct mitem_s
{
	struct mitem_s *next, *prev;
	size_t bytes;
	void *data;
} mitem_t;

typedef struct memtrack_s
{
	mitem_t		*head;
	mitem_t		*tail;
	int			count;
	int			bytes;
} memtrack_t;
*/

struct mlocal							// private data
{
	// Make this
	memchain_t	*memchain;	// Stores our memory allocations for easy disposal

	int			cleanup_flags;	
};

enum cleanup_flags_e { NET_STARTUP_1 = 1 };

///////////////////////////////////////////////////////////////////////////////
//  INTERNAL FUNCTIONS
///////////////////////////////////////////////////////////////////////////////

// Shutdown and Free
static cbool Listen (mobj_t *me)
{
	MLOCAL(m)
	
	// 10 is maximum unaccepted connections roughly
	if (listen (me->socket, 10) != 0)
	{
		Net_Error ("Listen");
		Core_Error ("Listen failed");
		return false;
	}

	return true;
}


///////////////////////////////////////////////////////////////////////////////
//  PUBLIC GLOBAL FUNCTIONS:  Creation and freeing of object
///////////////////////////////////////////////////////////////////////////////

// Shutdown and Free
static void *Shutdown (mobj_t *me)
{
	MLOCAL(m)

	if (m->cleanup_flags & NET_STARTUP_1)
	{
		Net_Cleanup ();
		m->cleanup_flags &= ~NET_STARTUP_1;
	}

	if (me->socket) 
	{
		sys_closesocket (me->socket);
		me->socket = INVALID_SOCKET;
	}


	m->memchain->Shutdown (m->memchain);

	core_free (me->_local);
	core_free (me);

	return NULL;
}



netserv_t *Netserv_TCP_Instance (const char *ip_address, int port)
{
	struct mlocal *m;

	#define INSTANCE_ERROR(_s) { error_step = _s; goto server_fail; }
	const char *error_step = NULL;
	
	mobj_t *nobj = core_calloc (sizeof(netserv_t), 1);
	mobj_t *me = nobj;
	me->_local = core_calloc (sizeof(struct mlocal), 1);
	m = (struct mlocal *) me->_local;

	m->memchain = Memchain_Instance ();

	me->ip_requested = m->memchain->Strdup (m->memchain, ip_address ? ip_address : "");
	me->port = port;

	me->socket = INVALID_SOCKET;
	me->ip_protocol = IPPROTO_TCP;
	
	// #1 - Net_Startup increments count.  Windows: does WSAStartup
	Net_Startup ();
	m->cleanup_flags |= NET_STARTUP_1; 
	
	// Setup socket storage
	// Note: Don't give sockaddr_storage_set a NULL ip address 
	// because can't determine IPv4 or IPv6 with no information: ...
	if (ip_address == NULL)
	{
		me->servaddr.ss_family = AF_INET6;
		INETADDR_SETANY((struct sockaddr *)&me->servaddr);
		SS_PORT(&me->servaddr) = htons((unsigned short)me->port);
	}
	else sockaddr_storage_set (&me->servaddr, ip_address, &port); 

//	sockaddr_storage_anyset6 (&me->servaddr, &port);

	sockaddr_storage_printf (&me->servaddr);

	me->socket = socket(me->servaddr.ss_family, SOCK_STREAM, IPPROTO_TCP);

	//make the socket a dual mode socket
	{
	int off = 0, on = 1;
	if (me->servaddr.ss_family == AF_INET6)
		setsockopt (me->socket, IPPROTO_IPV6, IPV6_V6ONLY, (char *)&off, sizeof(off));

#ifndef PLATFORM_WINDOWS
	if ((fcntl (me->socket, F_SETFL, O_NONBLOCK)) == -1)
		INSTANCE_ERROR ("fcntl nonblocking");
#endif

	if (ioctlsocket (me->socket, FIONBIO, &on) == -1)
		INSTANCE_ERROR ("ioctlsocket nonblocking");

	} //

	// A TCP/IP client might not want to bind, this locks us to a specific port
	if (bind (me->socket, (struct sockaddr *)&me->servaddr, sizeof me->servaddr) != 0)
		INSTANCE_ERROR ("bind");

	// Get the port --- we must have to do something like connect or sendto first
	sockaddr_storage_printf (&me->servaddr);

	// Hook up functions
	#define FUNCTION_HOOKUP(_x) nobj->_x = _x

	FUNCTION_HOOKUP(Listen);

	#undef FUNCTION_HOOKUP

	// Required information
	OBJ_REQUIRED_HOOKUP(nobj) // Sets parent, _tag, Shutdown

	return me;


server_fail:

	Net_Error (error_step);

	Shutdown (me);	

	return NULL;
}

